#!/usr/bin/env python

"""Python script to run cell model"""


"""
/* Copyright (c) 2015 EPFL-BBP, All rights reserved.

THIS SOFTWARE IS PROVIDED BY THE BLUE BRAIN PROJECT ``AS IS''
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE BLUE BRAIN PROJECT
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

This work is licensed under a
Creative Commons Attribution-NonCommercial-ShareAlike 4.0 International License.
To view a copy of this license, visit
http://creativecommons.org/licenses/by-nc-sa/4.0/legalcode or send a letter to
Creative Commons, 171 Second Street, Suite 300,
San Francisco, California, 94105, USA.
"""

"""
 * @file run.py
 * @brief Run simulation using pyneuron
 * @author Werner Van Geit @ BBP
 * @date 2015
"""

# pylint: disable=C0325, W0212, F0401, W0612, F0401

import os
import neuron
import numpy
import sys


def create_cell():
    """Create the cell model"""
    # Load morphology
    neuron.h.load_file("morphology.hoc")
    # Load biophysics
    neuron.h.load_file("biophysics.hoc")
    # Load main cell template
    neuron.h.load_file("template.hoc")

    # Instantiate the cell from the template

    print("Loading cell bNAC219_L5_SBC_77ab259b7f")
    cell = neuron.h.bNAC219_L5_SBC_77ab259b7f(0)
    return cell


def create_stimuli(cell, stim_start, stim_end, current_amplitude):
    """Create the stimuli"""

    print('Attaching stimulus electrodes')

    stimuli = []

    iclamp = neuron.h.IClamp(0.5, sec=cell.soma[0])
    iclamp.delay = stim_start
    iclamp.dur = stim_end - stim_start
    iclamp.amp = current_amplitude
    print('Setting up step current clamp: '
          'amp=%f nA, delay=%f ms, duration=%f ms' %
          (iclamp.amp, iclamp.delay, iclamp.dur))

    stimuli.append(iclamp)

    return stimuli


def create_recordings(cell):
    """Create the recordings"""
    print('Attaching recording electrodes')

    recordings = {}

    recordings['time'] = neuron.h.Vector()
    recordings['soma(0.5)'] = neuron.h.Vector()

    recordings['time'].record(neuron.h._ref_t, 0.1)
    recordings['soma(0.5)'].record(cell.soma[0](0.5)._ref_v, 0.1)

    return recordings


def run_RmpRiTau_step(
        stim_start,
        stim_end,
        current_amplitude,
        plot_traces=None):
    """Run """

    cell = create_cell()
    stimuli = create_stimuli(cell, stim_start, stim_end, current_amplitude)  # noqa
    recordings = create_recordings(cell)

    # Overriding default 30s simulation,
    neuron.h.tstop = stim_end + stim_start
    print(
        'Setting simulation time to %.6g ms for the step current' %
        neuron.h.tstop)

    print('Setting initial voltage to -70 mV')
    neuron.h.v_init = -70

    neuron.h.stdinit()
    neuron.h.dt = 1000
    neuron.h.t = -1e9
    for _ in range(10):
        neuron.h.fadvance()

    neuron.h.t = 0
    neuron.h.dt = 0.025
    neuron.h.frecord_init()

    neuron.h.continuerun(3000)

    time = numpy.array(recordings['time'])
    soma_voltage = numpy.array(recordings['soma(0.5)'])

    recordings_dir = 'python_recordings'

    soma_voltage_filename = os.path.join(
        recordings_dir,
        'soma_voltage_RmpRiTau_step.dat')
    numpy.savetxt(soma_voltage_filename, zip(time, soma_voltage))

    print('Soma voltage for RmpRiTau trace saved to: %s'
          % (soma_voltage_filename))

    if plot_traces:
        import pylab
        pylab.figure(facecolor='white')
        pylab.plot(recordings['time'], recordings['soma(0.5)'])
        pylab.xlabel('time (ms)')
        pylab.ylabel('Vm (mV)')
        pylab.gcf().canvas.set_window_title('RmpRiTau trace')

    return time, soma_voltage, stim_start, stim_end


def init_simulation():
    """Initialise simulation environment"""

    neuron.h.load_file("stdrun.hoc")
    neuron.h.load_file("import3d.hoc")

    print('Loading constants')
    neuron.h.load_file('constants.hoc')


def analyse_RmpRiTau_trace(
        time,
        soma_voltage,
        stim_start,
        stim_end,
        current_amplitude):
    """Analyse the output of the RmpRiTau protocol"""

    # Import the eFeature Extraction Library
    import efel

    # Prepare the trace data
    trace = {}
    trace['T'] = time
    trace['V'] = soma_voltage
    trace['stim_start'] = [stim_start]
    trace['stim_end'] = [stim_end]

    # Calculate the necessary eFeatures
    efel_results = efel.getFeatureValues(
        [trace],
        ['voltage_base', 'steady_state_voltage_stimend',
         'decay_time_constant_after_stim'])

    voltage_base = efel_results[0]['voltage_base'][0]
    ss_voltage = efel_results[0]['steady_state_voltage_stimend'][0]
    dct = efel_results[0]['decay_time_constant_after_stim'][0]

    # Calculate input resistance
    input_resistance = float(ss_voltage - voltage_base) / current_amplitude

    rmpritau_dict = {}

    rmpritau_dict['Rmp'] = '%.6g' % voltage_base
    rmpritau_dict['Rmp_Units'] = 'mV'
    rmpritau_dict['Rin'] = '%.6g' % input_resistance
    rmpritau_dict['Rin_Units'] = 'MOhm'
    rmpritau_dict['Tau'] = '%.6g' % dct
    rmpritau_dict['Tau_Units'] = 'ms'

    print('Resting membrane potential is %s %s' %
          (rmpritau_dict['Rmp'], rmpritau_dict['Rmp_Units']))
    print('Input resistance is %s %s' %
          (rmpritau_dict['Rin'], rmpritau_dict['Rin_Units']))
    print('Time constant is %s %s' %
          (rmpritau_dict['Tau'], rmpritau_dict['Tau_Units']))

    import json

    with open('rmp_ri_tau.json', 'w') as rmpritau_json_file:
        json.dump(rmpritau_dict, rmpritau_json_file,
                        sort_keys=True,
                        indent=4,
                        separators=(',', ': '))


def main(plot_traces=False):
    """Main"""

    # Import matplotlib to plot the traces
    if plot_traces:
        import matplotlib
        matplotlib.rcParams['path.simplify'] = False

    init_simulation()

    current_amplitude = -0.01
    stim_start = 1000
    stim_end = 2000

    time, soma_voltage, stim_start, stim_end = run_RmpRiTau_step(
        stim_start, stim_end, current_amplitude, plot_traces=plot_traces)

    analyse_RmpRiTau_trace(
        time,
        soma_voltage,
        stim_start,
        stim_end,
        current_amplitude)

    if plot_traces:
        import pylab
        pylab.show()

if __name__ == '__main__':
    if len(sys.argv) == 1:
        main(plot_traces=True)
    elif len(sys.argv) == 2 and sys.argv[1] == '--no-plots':
        main(plot_traces=False)
    else:
        raise Exception(
            "Script only accepts one argument: --no-plots, not %s" %
            str(sys.argv))



from MEArec.tools import load_templates
from MEArec.generators import RecordingGenerator, SpikeTrainGenerator, TemplateGenerator
import yaml
import os
from distutils.version import StrictVersion
import time
import numpy as np
import neo

if StrictVersion(yaml.__version__) >= StrictVersion('5.0.0'):
    use_loader = True
else:
    use_loader = False


def gen_recordings(params=None, templates=None, tempgen=None, spgen=None, verbose=True,
                   tmp_mode='memmap', tmp_folder=None, n_jobs=0):
    """
    Generates recordings.

    Parameters
    ----------
    templates : str
        Path to generated templates
    params : dict or str
        Dictionary containing recording parameters OR path to yaml file containing parameters
    tempgen : TemplateGenerator
        Template generator object
    spgen : SpikeTrainGenerator
        Spike train generator object. If None spike trains are created from params['spiketrains']
    verbose: bool or int
        Determines the level of verbose. If 1 or True, low-level, if 2 high level, if False, not verbose
    tmp_mode : None, 'h5' 'memmap'
        Use temporary file h5 memmap or None
        None is no temporary file.
    tmp_folder: str or Path
        In case of tmp files, you can specify the folder.
        If None, then it is automatic using tempfile.mkdtemp()

    Returns
    -------
    RecordingGenerator
        Generated recording generator object
    """
    t_start = time.perf_counter()
    if isinstance(params, str):
        if os.path.isfile(params) and (params.endswith('yaml') or params.endswith('yml')):
            with open(params, 'r') as pf:
                if use_loader:
                    params_dict = yaml.load(pf, Loader=yaml.FullLoader)
                else:
                    params_dict = yaml.load(pf)
    elif isinstance(params, dict):
        params_dict = params
    else:
        params_dict = {}

    if 'spiketrains' not in params_dict:
        params_dict['spiketrains'] = {}
    if 'templates' not in params_dict:
        params_dict['templates'] = {}
    if 'recordings' not in params_dict:
        params_dict['recordings'] = {}
    if 'cell_types' not in params_dict:
        params_dict['cell_types'] = {}
    if 'seeds' not in params_dict:
        params_dict['seeds'] = {}

    if tempgen is None and templates is None:
        raise AttributeError("Provide either 'templates' or 'tempgen' TemplateGenerator object")

    if tempgen is None:
        if os.path.isdir(templates):
            tempgen = load_templates(templates, verbose=False)
        elif templates.endswith('h5') or templates.endswith('hdf5'):
            tempgen = load_templates(templates, verbose=False)
        else:
            raise AttributeError("'templates' is not a folder or an hdf5 file")

    if 'spiketrains' in params_dict['seeds']:
        if params_dict['seeds']['spiketrains'] is None:
            params_dict['seeds']['spiketrains'] = np.random.randint(1, 10000)
    else:
        params_dict['seeds']['spiketrains'] = np.random.randint(1, 10000)

    if 'templates' in params_dict['seeds']:
        if params_dict['seeds']['templates'] is None:
            params_dict['seeds']['templates'] = np.random.randint(1, 10000)
    else:
        params_dict['seeds']['templates'] = np.random.randint(1, 10000)

    if 'convolution' in params_dict['seeds']:
        if params_dict['seeds']['convolution'] is None:
            params_dict['seeds']['convolution'] = np.random.randint(1, 10000)
    else:
        params_dict['seeds']['convolution'] = np.random.randint(1, 10000)

    if 'noise' in params_dict['seeds']:
        if params_dict['seeds']['noise'] is None:
            params_dict['seeds']['noise'] = np.random.randint(1, 10000)
    else:
        params_dict['seeds']['noise'] = np.random.randint(1, 10000)

    # Generate spike trains
    if spgen is None:
        spgen = SpikeTrainGenerator(params_dict['spiketrains'], verbose=verbose,
                                    seed=params_dict['seeds']['spiketrains'])
        spgen.generate_spikes()
    else:
        assert isinstance(spgen, SpikeTrainGenerator), "'spgen' should be a SpikeTrainGenerator object"
        spgen.info['custom'] = True

    params_dict['spiketrains'] = spgen.info
    # Generate recordings
    recgen = RecordingGenerator(spgen, tempgen, params_dict)
    recgen.generate_recordings(tmp_mode=tmp_mode, tmp_folder=tmp_folder, n_jobs=n_jobs, verbose=verbose)

    if verbose >= 1:
        print('Elapsed time: ', time.perf_counter() - t_start)

    return recgen


def gen_spiketrains(params=None, spiketrains=None, seed=None, verbose=False):
    """
    Generates spike trains.

    Parameters
    ----------
    params : str or dict
        Path to parameters yaml file or parameters dictionary
    spiketrains : list
        List of neo.SpikeTrains (alternative to params definition)
    verbose : bool
        If True, the output is verbose

    Returns
    -------
    SpikeTrainGenerator
        Generated spike train generator object
    """
    if params is None:
        assert spiketrains is not None, "Pass either a 'params' or a 'spiketrains' argument"
        assert isinstance(spiketrains, list) and isinstance(spiketrains[0], neo.SpikeTrain), \
            "'spiketrains' should be a list of neo.SpikeTrain objects"
        params_dict = {}
    else:
        if isinstance(params, str):
            if os.path.isfile(params) and (params.endswith('yaml') or params.endswith('yml')):
                with open(params, 'r') as pf:
                    if use_loader:
                        params_dict = yaml.load(pf, Loader=yaml.FullLoader)
                    else:
                        params_dict = yaml.load(pf)
        elif isinstance(params, dict):
            params_dict = params
        else:
            params_dict = {}
        spiketrains = None

    spgen = SpikeTrainGenerator(params=params_dict, spiketrains=spiketrains, seed=seed, verbose=verbose)
    spgen.generate_spikes()

    return spgen


def gen_templates(cell_models_folder, params=None, templates_tmp_folder=None,
                  intraonly=False, parallel=True, n_jobs=None, recompile=False, delete_tmp=True, verbose=True):
    """

    Parameters
    ----------
    cell_models_folder : str
        path to folder containing cell models
    params : str or dict
        Path to parameters yaml file or parameters dictionary
    templates_tmp_folder: str
        Path to temporary folder where templates are temporarily saved
    intraonly : bool
        if True, only intracellular simulation is run
    parallel : bool
        if True, multi-threading is used
    n_jobs: int
        Number of jobs to run in parallel (If None all cpus are used)
    recompile: bool
        If True, cell models are recompiled
    delete_tmp :
        if True, the temporary files are deleted
    verbose : bool
        If True, the output is verbose

    Returns
    -------
    TemplateGenerator
        Generated template generator object

    """
    if isinstance(params, str):
        if os.path.isfile(params) and (params.endswith('yaml') or params.endswith('yml')):
            with open(params, 'r') as pf:
                if use_loader:
                    params_dict = yaml.load(pf, Loader=yaml.FullLoader)
                else:
                    params_dict = yaml.load(pf)
    elif isinstance(params, dict):
        params_dict = params
    else:
        params_dict = None

    if templates_tmp_folder is not None:
        if not os.path.isdir(templates_tmp_folder):
            os.makedirs(templates_tmp_folder)

    tempgen = TemplateGenerator(cell_models_folder=cell_models_folder,
                                params=params_dict,
                                templates_folder=templates_tmp_folder,
                                intraonly=intraonly,
                                parallel=parallel,
                                recompile=recompile,
                                n_jobs=n_jobs,
                                delete_tmp=delete_tmp,
                                verbose=verbose)
    tempgen.generate_templates()

    return tempgen



"""
This module group several sub function for recording generator.

All this function work on chunk of signals.
They can be call in loop mode or with joblib.

Important:

When tmp_mode=='memmap' : theses functions must assign and add directly the buffer.
When tmp_mode is Noe : theses functions return the buffer and the assignament is done externally.



"""
import h5py
import numpy as np
import scipy.signal

from MEArec.tools import (filter_analog_signals, convolve_templates_spiketrains,
                          convolve_single_template, convolve_drifting_templates_spiketrains)


class FuncThenAddChunk:
    """
    Helper for functions that do chunk to assign one or several chunks at the 
    good place.
    """

    def __init__(self, func):
        self.func = func

    def __call__(self, *args, **kargs):
        return_dict = self.func(*args)

        ch, i_start, i_stop, = args[:3]

        assignment_dict = kargs['assignment_dict']
        tmp_mode = kargs['tmp_mode']

        if tmp_mode is None:
            pass
        elif tmp_mode == 'memmap':
            for key, full_arr in assignment_dict.items():
                out_chunk = return_dict.pop(key)
                full_arr[i_start:i_stop] += out_chunk

        return return_dict


def chunk_convolution_(ch, i_start, i_stop, chunk_start,
                       st_idxs, pad_samples, modulation, drifting, drift_mode, drifting_units, templates,
                       cut_outs_samples, template_locs, velocity_vector, fast_drift_period, fast_drift_min_jump,
                       fast_drift_max_jump, t_start_drift, fs, verbose, amp_mod, bursting_units, shape_mod,
                       shape_stretch, extract_spike_traces, voltage_peaks, dtype, seed_list):
    """
    Perform full convolution for all spike trains by chunk.

    Parameters
    ----------
    ch: int
        Chunk id
    i_start: int
        first index of chunk
    i_stop: 
        last index of chunk (exclude)
    chunk_start: quantity
        Start time for current chunk
    st_idxs: list
        list of spike indexes for each spike train
    pad_samples: int
        Number of samples for padding
    modulation: str
        Modulation type
    drifting: bool
        If True drifting is performed
    drift_mode :  str
        Drift mode ['slow' | 'fast' | 'slow+fast']
    drifting_units: list
        List of drifting units (if None all units are drifted)
    templates: np.array
        Templates
    cut_outs_samples: list
        List with number of samples to cut out before and after spike peaks
    template_locs: np.array
        For drifting, array with drifting locations
    velocity_vector: np.array
        For drifting, drifring direction
    fast_drift_period : Quantity
        Periods between fast drifts
    fast_drift_min_jump : float
        Min 'jump' in um for fast drifts
    fast_drift_max_jump : float
        Max 'jump' in um for fast drifts
    t_start_drift: quantity
        For drifting, start drift time
    fs: quantity
        Sampling frequency
    verbose: bool
        If True output is verbose
    amp_mod: np.array
        Array with modulation values
    bursting_units : list
        List of bursting units
    shape_mod: bool
        If True waveforms are modulated in shape
    shape_stretch: float
        Low and high frequency for bursting
    extract_spike_traces: bool
        If True (default), spike traces are extracted
    voltage_peaks: np.array
        Array containing the voltage values at the peak
    
    """
    if verbose:
        print('Start convolutions for chunk', ch)
    # set seed
    np.random.seed(seed_list[ch])

    # checking boundaries not needed here because the recordings are created on the fly
    i_start_pad = i_start - pad_samples
    i_stop_pad = i_stop + pad_samples
    n_samples = i_stop_pad - i_start_pad

    template_idxs = []
    if extract_spike_traces:
        spike_traces = np.zeros((n_samples, len(st_idxs)), dtype=dtype)
    if len(templates.shape) == 4:
        n_elec = templates.shape[2]
    elif len(templates.shape) == 5:
        n_elec = templates.shape[3]
    else:
        raise AttributeError("Wrong 'templates' shape!")

    recordings = np.zeros((n_samples, n_elec), dtype=dtype)

    for st, st_idx in enumerate(st_idxs):
        if extract_spike_traces:
            max_electrode = np.argmax(voltage_peaks[st])

        seed = np.random.randint(10000)
        np.random.seed(seed)

        if modulation in ['electrode', 'template']:
            mod_bool = True
            if bursting_units is not None:
                if st in bursting_units and shape_mod:
                    unit_burst = True
                else:
                    unit_burst = False
            else:
                unit_burst = False
            mod_array = amp_mod[st]
        else:  # modulation 'none'
            mod_bool = False
            mod_array = None
            unit_burst = False

        spike_idx_in_chunk_pad = np.where((st_idx > i_start_pad) & (st_idx <= i_stop_pad))
        spike_idx_in_chunk = np.where((st_idx > i_start) & (st_idx <= i_stop))

        if len(spike_idx_in_chunk_pad) > 0:
            st_idx_in_chunk_pad = st_idx[spike_idx_in_chunk_pad[0]] - i_start_pad

            if drifting and st in drifting_units:
                recordings, template_idx = convolve_drifting_templates_spiketrains(st, st_idx_in_chunk_pad,
                                                                                   templates[st],
                                                                                   n_samples=n_samples,
                                                                                   cut_out=cut_outs_samples,
                                                                                   modulation=mod_bool,
                                                                                   mod_array=mod_array,
                                                                                   fs=fs,
                                                                                   loc=template_locs[st],
                                                                                   drift_mode=drift_mode,
                                                                                   slow_drift_velocity=velocity_vector,
                                                                                   fast_drift_period=fast_drift_period,
                                                                                   fast_drift_min_jump=fast_drift_min_jump,
                                                                                   fast_drift_max_jump=fast_drift_max_jump,
                                                                                   t_start_drift=t_start_drift,
                                                                                   chunk_start=chunk_start,
                                                                                   bursting=unit_burst,
                                                                                   shape_stretch=shape_stretch,
                                                                                   verbose=verbose,
                                                                                   recordings=recordings)
                np.random.seed(seed)
                if extract_spike_traces:
                    spike_traces[:, st] = convolve_single_template(st, st_idx_in_chunk_pad,
                                                                   templates[st, 0, :, max_electrode],
                                                                   n_samples=n_samples,
                                                                   cut_out=cut_outs_samples,
                                                                   modulation=mod_bool,
                                                                   mod_array=mod_array,
                                                                   bursting=unit_burst,
                                                                   shape_stretch=shape_stretch)

                # only keep template idxs inside the chunk
                if len(spike_idx_in_chunk) > 0:
                    if len(spike_idx_in_chunk_pad[0]) != len(spike_idx_in_chunk[0]):
                        common_idxs = [i for i, idx in enumerate(spike_idx_in_chunk_pad[0])
                                       if idx in spike_idx_in_chunk[0]]
                        template_idx = template_idx[common_idxs]
                        assert len(template_idx) == len(spike_idx_in_chunk[0])
                else:
                    template_idx = np.array([])
            else:
                if drifting:
                    template = templates[st, 0]
                else:
                    template = templates[st]
                recordings = convolve_templates_spiketrains(st, st_idx_in_chunk_pad, template,
                                                            n_samples=n_samples,
                                                            cut_out=cut_outs_samples,
                                                            modulation=mod_bool,
                                                            mod_array=mod_array,
                                                            bursting=unit_burst,
                                                            shape_stretch=shape_stretch,
                                                            verbose=verbose,
                                                            recordings=recordings)
                np.random.seed(seed)
                if extract_spike_traces:
                    spike_traces[:, st] = convolve_single_template(st, st_idx_in_chunk_pad,
                                                                   template[:, max_electrode],
                                                                   n_samples=n_samples,
                                                                   cut_out=cut_outs_samples,
                                                                   modulation=mod_bool,
                                                                   mod_array=mod_array,
                                                                   bursting=unit_burst,
                                                                   shape_stretch=shape_stretch)
                template_idx = np.array([])
        else:
            if verbose:
                print('No spikes found in chunk', ch, 'for spike train', st)
            template_idx = np.array([])
        template_idxs.append(template_idx)

    if verbose:
        print('Done all convolutions for chunk', ch)

    return_dict = dict()
    if pad_samples > 0:
        return_dict['recordings'] = recordings[pad_samples:-pad_samples]
    else:
        return_dict['recordings'] = recordings
    if extract_spike_traces:
        return_dict['spike_traces'] = spike_traces[pad_samples:-pad_samples]
    return_dict['template_idxs'] = template_idxs

    return return_dict


chunk_convolution = FuncThenAddChunk(chunk_convolution_)


def chunk_uncorrelated_noise_(ch, i_start, i_stop, chunk_start,
                              num_chan, noise_level, noise_color, color_peak, color_q, color_noise_floor, fs, dtype,
                              seed_list):
    np.random.seed(seed_list[ch])
    length = i_stop - i_start
    additive_noise = noise_level * np.random.randn(length, num_chan).astype(dtype)

    if noise_color:
        # iir peak filter
        b_iir, a_iir = scipy.signal.iirpeak(color_peak, Q=color_q, fs=fs)
        additive_noise = scipy.signal.filtfilt(b_iir, a_iir, additive_noise, axis=0, padlen=1000)
        additive_noise = additive_noise.astype(dtype)
        additive_noise += color_noise_floor * np.std(additive_noise) * \
                          np.random.randn(additive_noise.shape[0], additive_noise.shape[1])
        additive_noise = additive_noise * (noise_level / np.std(additive_noise))

    return_dict = {}
    return_dict['additive_noise'] = additive_noise

    return return_dict


chunk_uncorrelated_noise = FuncThenAddChunk(chunk_uncorrelated_noise_)


def chunk_distance_correlated_noise_(ch, i_start, i_stop, chunk_start,
                                     noise_level, cov_dist, n_elec, noise_color, color_peak, color_q, color_noise_floor,
                                     fs, dtype, seed_list):
    np.random.seed(seed_list[ch])
    length = i_stop - i_start

    additive_noise = noise_level * np.random.multivariate_normal(np.zeros(n_elec), cov_dist,
                                                                 size=length).astype(dtype)
    if noise_color:
        # iir peak filter
        b_iir, a_iir = scipy.signal.iirpeak(color_peak, Q=color_q, fs=fs)
        additive_noise = scipy.signal.filtfilt(b_iir, a_iir, additive_noise, axis=0, padlen=1000)
        additive_noise = additive_noise + color_noise_floor * np.std(additive_noise) * \
                         np.random.multivariate_normal(np.zeros(n_elec), cov_dist,
                                                       size=length)
    additive_noise = additive_noise * (noise_level / np.std(additive_noise))

    return_dict = {}
    return_dict['additive_noise'] = additive_noise

    return return_dict


chunk_distance_correlated_noise = FuncThenAddChunk(chunk_distance_correlated_noise_)


def chunk_apply_filter_(ch, i_start, i_stop, chunk_start,
                        recordings, pad_samples, cutoff, order, fs, dtype):
    n_samples = recordings.shape[0]

    # compute padding idxs
    if pad_samples > 0:
        if i_start > pad_samples:
            i_start_pad = i_start - pad_samples
            pad_start_samples = pad_samples
        else:
            i_start_pad = 0
            pad_start_samples = i_start
        if n_samples - i_stop > pad_samples:
            i_stop_pad = i_stop + pad_samples
            pad_stop_samples = pad_samples
        else:
            i_stop_pad = n_samples
            pad_stop_samples = n_samples - i_stop
    else:
        i_start_pad = i_start
        i_stop_pad = i_stop

    if cutoff.size == 1:
        filtered_chunk = filter_analog_signals(recordings[i_start_pad:i_stop_pad], freq=cutoff, fs=fs,
                                               filter_type='highpass', order=order)
    elif cutoff.size == 2:
        if fs / 2. < cutoff[1]:
            filtered_chunk = filter_analog_signals(recordings[i_start_pad:i_stop_pad], freq=cutoff[0], fs=fs,
                                                   filter_type='highpass', order=order)
        else:
            filtered_chunk = filter_analog_signals(recordings[i_start_pad:i_stop_pad], freq=cutoff, fs=fs)

    filtered_chunk = filtered_chunk.astype(dtype)
    if pad_samples > 0:
        filtered_chunk = filtered_chunk[pad_start_samples:filtered_chunk.shape[0] - pad_stop_samples]

    return_dict = {}
    return_dict['filtered_chunk'] = filtered_chunk

    return return_dict


chunk_apply_filter = FuncThenAddChunk(chunk_apply_filter_)
